﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using Cysharp.Threading.Tasks;
using JetBrains.Annotations;
using UniRx;
using UnityEngine;
using static Magify.MagifyManager;

namespace Magify
{
    internal class SubscriptionService : IDisposable
    {
        private static readonly MagifyLogger _logger = MagifyLogger.Get(MagifyService.LogScope);

        [NotNull]
        private readonly ServicePrefs _prefs;
        [NotNull]
        private readonly ServiceTime _time;
        [NotNull]
        private readonly INetworkStatusProvider _network;
        [NotNull]
        private readonly IReadOnlyDictionary<CampaignType, CampaignRequest> _currentCampaigns;
        private readonly Subject<Unit> _onSubscriptionRefreshed = new Subject<Unit>();
        private readonly Subject<bool> _onSubscriptionTrialStatusRead = new Subject<bool>();

        private IInAppStore _inAppStore;
        private CancellationTokenSource _updateCancellationToken;
        private bool _firstLoadPassed;
        private bool _hotStartPassed;

        public IObservable<Unit> OnSubscriptionRefreshed => _onSubscriptionRefreshed;
        /// <summary>
        /// <b>True</b> if subscription is in trial period
        /// </summary>
        public IObservable<bool> OnSubscriptionTrialStatusRead => _onSubscriptionTrialStatusRead;

        internal SubscriptionService([NotNull] ServicePrefs prefs, [NotNull] ServiceTime time, [NotNull] INetworkStatusProvider network, [NotNull] IReadOnlyDictionary<CampaignType, CampaignRequest> currentCampaigns)
        {
            _prefs = prefs;
            _time = time;
            _network = network;
            _currentCampaigns = currentCampaigns;
        }

        public void SetPurchasingProvider([CanBeNull] IInAppStore inAppStore)
        {
            if (_inAppStore != null)
            {
                _inAppStore.OnPurchaseFinished -= PurchaseFinishedHandler;
                _inAppStore.OnProductFetched -= ProductFetchedHandler;
            }

            _inAppStore = inAppStore;
            if (_inAppStore != null)
            {
                _inAppStore.OnPurchaseFinished += PurchaseFinishedHandler;
                _inAppStore.OnProductFetched += ProductFetchedHandler;
            }
        }

        public void HandleAppFocus(bool hasFocus)
        {
            if (!hasFocus)
            {
                _hotStartPassed = false;
                return;
            }
            if (_inAppStore == null || _updateCancellationToken != null || _hotStartPassed || !_firstLoadPassed || _prefs.BoughtSubscriptionProduct.Value == null || !_network.IsNetworkReachable)
            {
                return;
            }
            _hotStartPassed = true;
            RefreshSubscription(_prefs.BoughtSubscriptionProduct.Value);
        }

        private void ProductFetchedHandler(string productId)
        {
            if (_updateCancellationToken != null || _prefs.BoughtSubscriptionProduct.Value != productId || _hotStartPassed)
            {
                return;
            }

            _firstLoadPassed = true;
            _hotStartPassed = true;
            _inAppStore.OnProductFetched -= ProductFetchedHandler;
            RefreshSubscription(productId);
        }

        internal void RefreshSubscription(string productId)
        {
            var product = SubscriptionProducts.FirstOrDefault(c => c.Id == productId);
            if (product == null)
            {
                _logger.LogError($"Unexpected subscription product! {_prefs.BoughtSubscriptionProduct.Value} bought but is not in {nameof(SubscriptionProducts)}");
                _prefs.BoughtSubscriptionProduct.Value = null;
                _prefs.SubscriptionExpireDate.Value = DateTime.UnixEpoch;
                _prefs.SubscriptionStatus.Value = SubscriptionStatus.Inactive;
                return;
            }

            var info = _inAppStore.LoadSubscriptionInfo(productId);
            if (info == null)
            {
                markAsExpired();
            }
            else if (_time.IsNetworkTimeProtected.Value || Application.platform == RuntimePlatform.Android)
            {
                updateCurrentSubscription();
            }
            else
            {
                waitServerTimeAndUpdate().Forget();
            }

            async UniTaskVoid waitServerTimeAndUpdate() // ios specific logic
            {
                _updateCancellationToken = new CancellationTokenSource();
                try
                {
                    await UniTask.WaitUntil(() => _time.IsNetworkTimeProtected.Value, cancellationToken: _updateCancellationToken.Token);
                    _updateCancellationToken.Cancel();
                    _updateCancellationToken.Dispose();
                    _updateCancellationToken = null;
                }
                catch (OperationCanceledException)
                {
                }
                if (info.IsExpired(_time.UtcNow))
                {
                    markAsExpired();
                }
            }

            void updateCurrentSubscription()
            {
                _prefs.SubscriptionExpireDate.Value = info.ExpireDate;
                _prefs.SubscriptionStatus.Value = GetSubscriptionInfoStatus(info.IsFreeTrial, false);
                _onSubscriptionRefreshed.OnNext(Unit.Default);
                _onSubscriptionTrialStatusRead.OnNext(info.IsFreeTrial);
            }

            void markAsExpired()
            {
                _prefs.SubscriptionStatus.Value = _prefs.SubscriptionStatus.Value switch
                {
                    SubscriptionStatus.Paid or SubscriptionStatus.ExpiredPaid => SubscriptionStatus.ExpiredPaid,
                    SubscriptionStatus.Trial or SubscriptionStatus.ExpiredTrial => SubscriptionStatus.ExpiredTrial,
                    _ => SubscriptionStatus.Inactive,
                };
                _onSubscriptionRefreshed.OnNext(Unit.Default);
            }
        }

        private void PurchaseFinishedHandler(string productId, PurchaseInfo purchaseInfo)
        {
            var product = SubscriptionProducts.FirstOrDefault(c => c.Id == productId);
            if (product == null) return;
            if (purchaseInfo.Receipt == null || purchaseInfo.SubscriptionInfo == null)
            {
                _logger.LogError($"Saving new subscription failed. The requested product ({productId}) not purchased yet (receipt is empty).");
                return;
            }
            _logger.Log($"Saving new subscription: \nProduct id - {product.Id}, \nReceipt: {purchaseInfo.Receipt}");
            var info = purchaseInfo.SubscriptionInfo;
            if (_currentCampaigns.Values.Any(request => request.Campaign.IsCampaignWithCreativeAndProduct(productId)))
            {
                _logger.Log($"Campaign for {product.Id} is active. Tracking subscription activation.");
                TrackSubscriptionActivation(
                    info.IsFreeTrial,
                    product.Id,
                    purchaseInfo.Price,
                    purchaseInfo.Currency,
                    info.PeriodModel.ToString(),
                    purchaseInfo.TransactionId,
                    purchaseInfo.PurchaseToken,
                    null, // ToDo: add originalTransactionId sending
                    purchaseInfo.Receipt.Payload);
            }
            else
            {
                _logger.Log($"Campaign for {product.Id} is not active. Tracking external subscription activation.");
                TrackExternalSubscriptionActivation(
                    info.IsFreeTrial,
                    product.Id,
                    purchaseInfo.Price,
                    purchaseInfo.Currency,
                    purchaseInfo.TransactionId,
                    purchaseInfo.PurchaseToken,
                    null, // ToDo: add originalTransactionId sending
                    purchaseInfo.Receipt.Payload);
            }
            _prefs.BoughtSubscriptionProduct.Value = product.Id;
            _prefs.SubscriptionExpireDate.Value = info.ExpireDate;
            _prefs.SubscriptionBoughtDate.Value = DateTime.UtcNow;
            _prefs.SubscriptionStatus.Value = GetSubscriptionInfoStatus(info.IsFreeTrial, info.IsExpired(_time.UtcNow));
            _firstLoadPassed = true;
            _onSubscriptionRefreshed.OnNext(Unit.Default);
            _updateCancellationToken?.Cancel();
            _updateCancellationToken?.Dispose();
            _updateCancellationToken = null;
        }

        private static SubscriptionStatus GetSubscriptionInfoStatus(bool isTrial, bool isExpired)
        {
            return isTrial switch
            {
                true when isExpired => SubscriptionStatus.ExpiredTrial,
                true => SubscriptionStatus.Trial,
                false when isExpired => SubscriptionStatus.ExpiredPaid,
                false => SubscriptionStatus.Paid
            };
        }

        void IDisposable.Dispose()
        {
            if (_inAppStore != null)
            {
                _inAppStore.OnPurchaseFinished -= PurchaseFinishedHandler;
                _inAppStore.OnProductFetched -= ProductFetchedHandler;
            }
        }
    }
}