using System;
using System.Collections.Generic;
using Magify.Types;
using UniRx;
using UnityEngine;

namespace Magify
{
    internal class GeneralPrefs : ILegacyMigrator, IInitializable, IDisposable
    {
        private const string KeyMigratedUser = "migrated_user";

        private const string KeyClientId = "client_id";
        private const string KeyFirstInstalledVersion = "first_version";
        private const string KeyFirstLaunchDate = "first_launch";

        private const string KeySessionCounter = "session_counter";
        private const string KeyVersionSessionCounter = "version_session_counter";
        private const string KeySessionId = "session_id";
        private const string KeySessionOpenTime = "session_open_time";
        private const string KeySessionLastTime = "session_last_time";

        private const string KeyAuthToken = "auth_token";
        private const string KeyAuthTokenExpired = "auth_token_expired";

        private const string KeyLanguageCode = "language_code";
        private const string KeyCountryCode = "country_code";

        private const string KeySubscriptionState = "subscription_state";
        private const string KeyInAppStatus = "in_app_status";
        private const string KeyAuthorizationStatus = "authorization_status";
        private const string KeyReferrerId = "referrer_id";
        private const string KeyAdjustId = "adjust_id";
        private const string KeyGpsAdId = "gps_ad_id";
        private const string KeyGdprStatus = "gdpr_granted";
        private const string KeyAttAuthorizationStatus = "att_authorization_status";

        private const string KeySourceNetwork = "source_network";
        private const string KeySourceCampaign = "source_campaign";
        private const string KeySourceGroup = "source_group";

        private const string KeyDefaultCampaignTypes = "default_campaign_types";
        private const string KeyProductsReceiptTime = "products_by_dates";

        private const string KeyPurchasedInApps = "purch_products";
        private const string KeyPurchasedSubscriptions = "purch_subs_products";
        private const string KeyUsedFreeProducts = "free_products";

        private const string KeyTrackedAppInstalls = "tracked_app_installs";

        private const string KeyFirebaseInstanceID = "firebase_instance_id";

        private const string KeyNextDailyResetTime = "next_daily_reset_time";

        internal const SubscriptionStatus DefaultSubscriptionStatus = Magify.SubscriptionStatus.Inactive;
        internal const InAppStatus DefaultInAppStatus = Magify.InAppStatus.NeverPurchased;
        internal const AuthorizationStatus DefaultAuthorizationStatus = Magify.AuthorizationStatus.Unknown;

        private readonly BinaryStorage _storage;
        private readonly PlatformAPI _platform;
        private readonly IDictionary<string, int> _versionSessionCounter;

        public bool MigratedUser
        {
            get => _storage.Get<bool>(KeyMigratedUser);
            private set => _storage.Set(KeyMigratedUser, value);
        }

        public string ClientId
        {
            get => _storage.Get<string>(KeyClientId);
            set => _storage.Set(KeyClientId, value);
        }

        public string FirstInstalledVersion
        {
            get => _storage.Get<string>(KeyFirstInstalledVersion);
            set => _storage.Set(KeyFirstInstalledVersion, value);
        }

        public DateTime FirstLaunchDate
        {
            get => _storage.Get(KeyFirstLaunchDate, default(DateTime));
            set => _storage.Set(KeyFirstLaunchDate, value);
        }

        public bool IsNewUser => FirstInstalledVersion == Application.version;
        public IReactiveProperty<bool> AuthorizationTokenExpired { get; }
        public IReactiveProperty<string> AuthorizationToken { get; }
        public IReactiveProperty<string> LanguageCode { get; }
        public IReactiveProperty<string> CountryCode { get; }
        public IReactiveProperty<SubscriptionStatus> SubscriptionStatus { get; }
        public IReactiveProperty<InAppStatus> InAppStatus { get; }
        public IReactiveProperty<AuthorizationStatus> AuthorizationStatus { get; }
        public IReactiveProperty<string> ReferrerId { get; }
        public IReactiveProperty<string> AdjustId { get; }
        public IReactiveProperty<string> GpsAdId { get; }
        public IReactiveProperty<bool> IsGdprGranted { get; }
        public IReactiveProperty<bool?> AttAuthorizationStatus { get; }
        public ISet<string> UsedDefaultCampaignTypes { get; }
        public IDictionary<string, DateTime> ProductsReceiptTimestamps { get; }

        public ISet<string> PurchasedInAppProducts { get; }
        public ISet<string> PurchasedSubscriptionProducts { get; }
        public ISet<string> UsedFreeProducts { get; }
        public ICollection<string> AllUsedProducts { get; }

        public ISet<string> TrackedAppInstalls { get; }

        public IReactiveProperty<string> FirebaseInstanceId { get; }

        public IReactiveProperty<int> GlobalSessionCounter { get; }

        public IReactiveProperty<string> SessionId { get; }

        public int VersionSessionCounter
        {
            get => _versionSessionCounter[Application.version];
            set => _versionSessionCounter[Application.version] = value;
        }

        public long SessionStarted
        {
            get => _storage.Get(KeySessionOpenTime, 0L);
            set => _storage.Set(KeySessionOpenTime, value);
        }

        public long LastActivity
        {
            get => _storage.Get(KeySessionLastTime, 0L);
            set => _storage.Set(KeySessionLastTime, value);
        }

        public MediaSource MediaSource
        {
            get
            {
                var network = _storage.Get<string>(KeySourceNetwork);
                var campaign = _storage.Get<string>(KeySourceCampaign);
                var group = _storage.Get<string>(KeySourceGroup);
                return new MediaSource(network, campaign, group);
            }
            set
            {
                using var _ = _storage.MultipleChangeScope();
                _storage.Set(KeySourceNetwork, value.NetworkName);
                _storage.Set(KeySourceCampaign, value.CampaignName);
                _storage.Set(KeySourceGroup, value.GroupName);
            }
        }

        public DateTime NextDailyResetTime
        {
            get => _storage.Get(KeyNextDailyResetTime, default(DateTime));
            set => _storage.Set(KeyNextDailyResetTime, value);
        }

        public static GeneralPrefs Create(string storagePath, PlatformAPI platform)
        {
            var storage = BinaryStorage
                .Construct(storagePath)
                .AddPrimitiveTypes()
                .SupportNullable<bool>()
                .SupportSetsOf<string>()
                .SupportDictionariesOf<string, long>()
                .SupportDictionariesOf<string, int>()
                .SupportDictionariesOf<string, DateTime>()
                .SupportEnum<SubscriptionStatus>()
                .SupportEnum<InAppStatus>()
                .SupportEnum<AuthorizationStatus>()
                .Build();

            var prefs = new GeneralPrefs(storage, platform);
            return prefs;
        }

        void ILegacyMigrator.Migrate(MigrationData migrationData)
        {
            using (MultipleChangeScope())
            {
                MigratedUser = true;
                ClientId = migrationData.ClientId;
                FirstInstalledVersion = migrationData.FirstInstalledVersion;
                FirstLaunchDate = migrationData.FirstLaunchDate ?? default;
                if (migrationData.SubscriptionStatus.HasValue)
                    SubscriptionStatus.Value = migrationData.SubscriptionStatus.Value;
                InAppStatus.Value = migrationData.InAppStatus ?? DefaultInAppStatus;
                if (migrationData.AuthorizationStatus.HasValue)
                    AuthorizationStatus.Value = migrationData.AuthorizationStatus.Value;
                ReferrerId.Value = migrationData.ReferrerId;
                AdjustId.Value = migrationData.AdjustId;
                GpsAdId.Value = migrationData.GpsAdId;
                IsGdprGranted.Value = migrationData.IsGdprGranted ?? default;
                AttAuthorizationStatus.Value = migrationData.AttAuthorizationStatus ?? default;
                UsedDefaultCampaignTypes.AddRange(migrationData.UsedDefaultCampaignTypes);
                ProductsReceiptTimestamps.AddRange(migrationData.ProductsReceiptTimestamps);
                PurchasedInAppProducts.AddRange(migrationData.PurchasedInAppProducts);
                PurchasedSubscriptionProducts.AddRange(migrationData.PurchasedSubscriptionProducts);
                UsedFreeProducts.AddRange(migrationData.UsedFreeProducts);
                TrackedAppInstalls.AddRange(migrationData.TrackedAppInstalls);

                FirebaseInstanceId.Value = migrationData.FirebaseInstanceId;
                GlobalSessionCounter.Value = migrationData.GlobalSessionCounter ?? default;
                SessionId.Value = migrationData.SessionId;
                SessionStarted = migrationData.SessionStarted ?? default;
                LastActivity = migrationData.LastActivity ?? default;
                MediaSource = migrationData.MediaSource ?? default;
            }
        }

        void IInitializable.Initialize()
        {
            if (ClientId == null)
            {
                var keychainValue = _platform.GetValueInNativeStorage(KeyClientId);
                if (keychainValue == null)
                {
                    ClientId = keychainValue = Guid.NewGuid().ToString();
                    _platform.SetValueInNativeStorage(keychainValue, KeyClientId);
                }
                else
                {
                    ClientId = keychainValue;
                }
            }
            if (string.IsNullOrEmpty(FirstInstalledVersion))
            {
                FirstInstalledVersion = Application.version;
            }
            if (FirstLaunchDate == default)
            {
                FirstLaunchDate = DateTime.Now;
            }
        }

        void IDisposable.Dispose()
        {
            _storage?.Dispose();
        }

        private GeneralPrefs(BinaryStorage storage, PlatformAPI platform)
        {
            _storage = storage;
            _platform = platform;
            using (MultipleChangeScope())
            {
                AuthorizationTokenExpired = _storage.GetReactiveProperty(KeyAuthTokenExpired, false);
                AuthorizationToken = _storage.GetReactiveProperty(KeyAuthToken, string.Empty);
                LanguageCode = _storage.GetReactiveProperty(KeyLanguageCode, string.Empty);
                CountryCode = _storage.GetReactiveProperty(KeyCountryCode, string.Empty);
                GlobalSessionCounter = _storage.GetReactiveProperty<int>(KeySessionCounter, 0);
                SessionId = _storage.GetReactiveProperty<string>(KeySessionId, string.Empty);
                SubscriptionStatus = _storage.GetReactiveProperty(KeySubscriptionState, DefaultSubscriptionStatus);
                InAppStatus = _storage.GetReactiveProperty(KeyInAppStatus, DefaultInAppStatus);
                AuthorizationStatus = _storage.GetReactiveProperty(KeyAuthorizationStatus, DefaultAuthorizationStatus);
                ReferrerId = _storage.GetReactiveProperty(KeyReferrerId, string.Empty);
                AdjustId = _storage.GetReactiveProperty(KeyAdjustId, default(string));
                GpsAdId = _storage.GetReactiveProperty<string>(KeyGpsAdId);
                IsGdprGranted = _storage.GetReactiveProperty<bool>(KeyGdprStatus);
                AttAuthorizationStatus = _storage.GetReactiveProperty<bool?>(KeyAttAuthorizationStatus);

                UsedDefaultCampaignTypes = _storage.GetSet<string>(KeyDefaultCampaignTypes);
                ProductsReceiptTimestamps = _storage.GetDictionary<string, DateTime>(KeyProductsReceiptTime);

                PurchasedInAppProducts = _storage.GetSet<string>(KeyPurchasedInApps);
                PurchasedSubscriptionProducts = _storage.GetSet<string>(KeyPurchasedSubscriptions);
                UsedFreeProducts = _storage.GetSet<string>(KeyUsedFreeProducts);
                AllUsedProducts = new UnionReadOnlyCollection<string>(PurchasedInAppProducts, PurchasedSubscriptionProducts, UsedFreeProducts);

                TrackedAppInstalls = _storage.GetSet<string>(KeyTrackedAppInstalls);
                FirebaseInstanceId = _storage.GetReactiveProperty<string>(KeyFirebaseInstanceID);

                _versionSessionCounter = _storage.GetDictionary<string, int>(KeyVersionSessionCounter);
                _versionSessionCounter.TryAdd(Application.version, 0);
            }
        }

        /// <summary>
        /// If you are planning to change many properties of this class, you should use this method.
        /// This method temporarily blocks the writing prefs file to disk inside <b>using</b> construction.
        /// </summary>
        /// <example>
        /// <code>
        /// using (MultipleChangeScope())
        /// {
        ///     GeneralPrefs.ClientId = Guid.NewGuid().ToString();
        ///     GeneralPrefs.AuthorizationStatus.Value = Magify.AuthorizationStatus.Authorized;
        /// }
        /// </code>
        /// </example>
        /// <returns>IDisposable which must disposed automatically after <b>using</b> construction (preferred) or disposed manually.</returns>
        public IDisposable MultipleChangeScope()
        {
            return _storage.MultipleChangeScope();
        }

        public void Reset()
        {
            _storage.ResetAll();
            _platform.RemoveValueInNativeStorage(KeyClientId);
        }
    }
}